import requests

from ...utilities import get_json_or_error


def trigger_miqa_test(miqa_server, trigger_id, version_name, api_key=None):
    """
      Trigger an offline execution of the specified Miqa trigger ID for the specified version.

      :param miqa_server: Miqa server e.g. yourco.miqa.io
      :param trigger_id: Miqa test trigger ID e.g. ABC123
      :param version_name: Name representing the version that was being tested/generated the outputs being uploaded, e.g. a commit ID ('abc123') or version number ('v1.2.3')
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: The ID for the Test Chain Run created as a result of this trigger (int).
      :rtype: int
    """

    trigger_url = f"https://{miqa_server}/api/test_trigger/{trigger_id}/execute?app=mn&offline_version=True&name={version_name}"
    headers = {}
    if api_key:
        headers['app_key'] = api_key

    trigger_response_json = get_json_or_error(trigger_url, headers=headers)
    run_id = trigger_response_json.get('run_id')
    if run_id:
        run_id = int(run_id)
    return run_id


def get_tcr_info_json(miqa_server, run_id, directory_containing_outputs=None, ds_id=None, wfv_id=None, api_key=None):
    """
      Retrieve a JSON describing the test chain run.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int run_id: Miqa Test Chain Run ID
      :param str directory_containing_outputs: Source location - local directory containing outputs to upload - only used if attempting to retrieve the upload command to use. (Optional)
      :param int ds_id: Datasource ID - enables getting specific information for the particular datasource
      :param int wfv_id: Workflow Variant ID - enables getting specific information for the particular workflow variant (i.e. if running a test on multiple workflow variants)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: JSON representation of Test Chain Run information
      :rtype: dict
    """
    get_info_url = f"https://{miqa_server}/api/get_tcr_exec_info/{run_id}"
    query_pars = []
    if directory_containing_outputs:
        query_pars.append(f"source_location={directory_containing_outputs}")
    if ds_id:
        query_pars.append(f"ds_id={ds_id}")
    if wfv_id:
        query_pars.append(f"&wfv_id={wfv_id}")

    if len(query_pars)>0:
        get_info_url = f"{get_info_url}?{'&'.join(query_pars)}"

    headers = {}
    if api_key:
        headers['app_key'] = api_key

    get_info_response_json = get_json_or_error(get_info_url, headers=headers)
    return get_info_response_json


def get_trigger_info(miqa_server, trigger_id, api_key=None):
    trigger_url = f"https://{miqa_server}/api/test_trigger/{trigger_id}/get_ds_id_mapping"

    headers = {}
    if api_key:
        headers['app_key'] = api_key

    trigger_response_json = get_json_or_error(trigger_url, headers=headers)
    # return {"ds_id_mapping":{"results":trigger_response_json, "url":trigger_url}}
    return {"ds_id_mapping":{"results":trigger_response_json}}


def get_execution_start_time_url(miqa_server, exec_id):
    return f"https://{miqa_server}/api/execution/{exec_id}/set_start_time"


def update_execution_start_time(miqa_server, exec_id, api_key, quiet=True):
    headers = {}
    if api_key:
        headers['app_key'] = api_key

    try:
        get_info_response_json = get_json_or_error(get_execution_start_time_url(miqa_server, exec_id), headers=headers, suppress_error_text=True)
        return get_info_response_json
    except Exception as e:
        if not quiet:
            print(f"Failed to update execution start time for {exec_id}")
            print(e)


import asyncio
import os
import re
import threading
import time

import aiohttp
import requests
from aiohttp import FormData

from ..baseuploadhelpers import get_remote_url, cloud_upload, get_user_log_post_url


async def upload_folder_async(folder, server, bucket, subfolder="folder-up3", filepattern=None, filepattern_start=None, quiet=True, cloud_provider='google', org_config_id=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, filepatterns=None, per_request_timeout=None, session=None, parallel_file_upload=True, detailed_file_logs=False, req_metadata=None):
    start_folder = time.perf_counter()
    if not quiet:
        print(f"Uploading folder {folder}")
    files = [os.path.join(dp, f) for dp, dn, fn in os.walk(os.path.expanduser(folder)) for f in fn if
             not f.endswith(".DS_Store")]
    max_files = len(files)
    sorted_files = sorted({k:os.stat(k).st_size for k in files if (not max_filesize or (os.stat(k).st_size/(1024*1024) < max_filesize))}.items(), key=lambda kv: kv[1])
    filesize_msg = f'(filtered out {(max_files - len(sorted_files))} by size)' if max_filesize else ''
    if not quiet:
        print(f"Processing up to {len(sorted_files)} files from this directory {filesize_msg}...")

    if not session:
        session_timeout = aiohttp.ClientTimeout(total=per_request_timeout)
        if max_connections:
            connector = aiohttp.TCPConnector(limit=max_connections)
            client = aiohttp.ClientSession(connector=connector, timeout=session_timeout)
        else:
            client = aiohttp.ClientSession(timeout=session_timeout)
    else:
        client = session
    async with client as session:
        tasks = []

        num_processed = 0
        num_skipped = 0
        file_times_miqa = {}
        file_times_cloud = {}
        filesize_lookup = {}

        for file, filesize in sorted_files:
            if (not filepatterns or check_any_pattern_matches(filepatterns, file)) and (not filepattern or file.endswith(filepattern)) and (not filepattern_start or file.startswith(filepattern_start)) and (not exclude_filepattern_end or not file.endswith(exclude_filepattern_end)):
                if not quiet and detailed_file_logs:
                    print(f"{file}: {round(filesize/ (1024 * 1024),5)}MB")
                filesize_lookup[os.path.basename(file)] = f"{round(filesize/ (1024 * 1024),5)}MB"
                subsubfolder = "/".join(file[len(folder)+1:].split("/")[:-1])
                filepath = os.path.join(folder, file)
                if os.path.isdir(filepath):
                    num_skipped+=1
                    if not quiet:
                        print(f"Skipping directory: {num_skipped} total skipped so far")
                    continue
                num_processed += 1

                start_miqaremote = time.perf_counter()
                remote_url = get_remote_url(bucket, filepath, server, subfolder+"/"+subsubfolder, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)
                end_miqaremote = time.perf_counter()
                time_miqaremote = end_miqaremote - start_miqaremote
                file_times_miqa[filepath] = time_miqaremote

                headers = {}
                if api_key:
                    headers['app_key'] = api_key
                async with session.get(remote_url, headers=headers) as response:
                    if not quiet and detailed_file_logs:
                        print(f"Threads: {threading.active_count()}")
                        print("Status:", response.status)
                        print("Content-type:", response.headers['content-type'])

                    if response.status != 200:
                        print(f"ERROR: Unable to post file: response status was {response.status}")
                        json_r = await response.json()
                        if json_r and json_r.get('message'):
                            print(json_r.get('message'))
                        print("------------------------------------")
                        return

                    json_r = await response.json()
                    # if not quiet and detailed_file_logs:
                    #     print(json_r)
                    new_url = json_r.get('url')
                    content_length = os.path.getsize(filepath)
                    payload = open(filepath, "rb")
                    headers_file = {
                        'Content-Length': f'{content_length}',
                        'Content-Type': 'text/plain'
                    }

                    start_cloudupload = time.perf_counter()
                    if cloud_provider == 'google':
                        if parallel_file_upload:
                            tasks.append(upload_to_gcp(filepath, headers_file, new_url, payload, quiet, session, detailed_file_logs))
                        else:
                            await upload_to_gcp(filepath, headers_file, new_url, payload, quiet, session, detailed_file_logs)
                    else:
                        if parallel_file_upload:
                            tasks.append(upload_to_aws(filepath, new_url, quiet, session, detailed_file_logs))
                        else:
                            await upload_to_aws(filepath, new_url, quiet, session, detailed_file_logs)
                    if not parallel_file_upload:
                        end_cloudupload = time.perf_counter()
                        time_cloudupload = end_cloudupload - start_cloudupload
                        file_times_cloud[filepath] = time_cloudupload
                        if not quiet and detailed_file_logs:
                            print(f"Time to upload {filepath}: {round(time_cloudupload + time_miqaremote,4)}s")

        if parallel_file_upload:
            end_prep = time.perf_counter()
            await asyncio.gather(*tasks) # execute them in concurrent manner
        end_folder = time.perf_counter()
        if not quiet:
            print(f"Total time: {end_folder - start_folder}s")
            if not parallel_file_upload:
                total_cloud_time = sum(file_times_cloud.values())
                total_miqa_time = sum(file_times_miqa.values())
                print(f"Done uploading for {num_processed} file(s). Total time: {round(total_cloud_time+total_miqa_time, 4)}s. {round(total_miqa_time, 4)}s on Miqa requests and {round(total_cloud_time, 4)}s on cloud uploads (Per-File Average {round(total_miqa_time/len(file_times_miqa), 4)}s Miqa / {round(total_cloud_time/len(file_times_cloud), 4)}s Cloud ; Max {round(max(file_times_miqa.values()), 4)}s Miqa / {round(max(file_times_cloud.values()), 4)}s).")
        if req_metadata:
            data = {}
            data.update(req_metadata)
            data["filesizes"] = filesize_lookup
            if file_times_cloud:
                data["filetimes"] = file_times_cloud
            data["total_time"] = end_folder - start_folder
            try:
                response = requests.post(get_user_log_post_url(server, req_metadata.get('tcr_id')), json=data, headers=headers)
                if not quiet:
                    print(f"Posted summary to Miqa with status {response.status_code}")
            except Exception as e:
                if not quiet:
                    print("Unable to post summary to Miqa")
                    print(e)


async def upload_to_gcp(filepath, headers, new_url, payload, quiet, session, detailed_file_logs=False):
    try:
        async with session.put(new_url, data=payload, headers=headers) as response:
            if not quiet and detailed_file_logs:
                print(f"submitted {filepath}")
                print(response)
    except Exception as e:
        if not quiet:
            print(f"Retrying upload for {filepath} after error")
            print(e)
        async with session.put(new_url, data=payload, headers=headers) as response:
            if not quiet and detailed_file_logs:
                print(f"submitted {filepath}")
                print(response)


async def upload_to_aws(filepath, new_url, quiet, session, detailed_file_logs=False):
    data = FormData()
    for dk, dv in new_url.get('fields', {}).items():
        data.add_field(dk, dv)
    data.add_field('file', open(filepath, 'rb'))
    try:
        async with session.post(new_url.get('url'), data=data) as response:
            if not quiet and detailed_file_logs:
                print(f"submitted {filepath}")
    except Exception as e:
        if not quiet:
            print(f"Retrying upload for {filepath} after error")
            print(e)
        async with session.post(new_url.get('url'), data=data) as response:
            if not quiet and detailed_file_logs:
                print(f"submitted {filepath}")



def check_any_pattern_matches(filepatterns, file):
    for filepattern in filepatterns:
        if check_pattern_match(filepattern, file):
            return True
    return False


def upload_folder_sync(folder, server, bucket, subfolder="folder-up3", filepattern=None, filepattern_start=None, quiet=True, cloud_provider='google', org_config_id=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, filepatterns=None):
    if not quiet:
        print("Processing folder synchronously...")
        print(f"Uploading folder {folder}")
    files = [os.path.join(dp, f) for dp, dn, fn in os.walk(os.path.expanduser(folder)) for f in fn if
             not f.endswith(".DS_Store")]
    max_files = len(files)
    sorted_files = sorted({k:os.stat(k).st_size for k in files if (not max_filesize or (os.stat(k).st_size/(1024*1024) < max_filesize))}.items(), key=lambda kv: kv[1])
    filesize_msg = f'(filtered out {(max_files - len(sorted_files))} by size)' if max_filesize else ''
    if not quiet:
        print(f"Processing up to {len(sorted_files)} files from this directory {filesize_msg}...")

    num_processed = 0
    num_skipped = 0
    for file, filesize in sorted_files:
        if (not filepatterns or check_any_pattern_matches(filepatterns, file)) and (not filepattern or file.endswith(filepattern)) and (not filepattern_start or file.startswith(filepattern_start)) and (not exclude_filepattern_end or not file.endswith(exclude_filepattern_end)):
            if not quiet:
                print(f"{file}: {round(filesize/ (1024 * 1024),3)}MB")
            subsubfolder = "/".join(file[len(folder)+1:].split("/")[:-1])
            filepath = os.path.join(folder, file)
            if os.path.isdir(filepath):
                num_skipped+=1
                if not quiet:
                    print(f"Skipping directory: {num_skipped} total skipped so far")
                continue
            num_processed += 1
            remote_url = get_remote_url(bucket, filepath, server, subfolder+"/"+subsubfolder, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)
            headers = {}
            if api_key:
                headers['app_key'] = api_key
            response = requests.get(remote_url, headers=headers)
            if not quiet:
                print("Status:", response.status_code)
                print("Content-type:", response.headers['content-type'])

            if response.status_code != 200:
                print(f"ERROR: Unable to post file: response status was {response.status_code}")
                json_r = response.json()
                if json_r and json_r.get('message'):
                    print(json_r.get('message'))
                print("------------------------------------")
                return

            json_r = response.json()
            if not quiet:
                print(json_r)
            new_url = json_r.get('url')
            content_length = os.path.getsize(filepath)
            payload = open(filepath, "rb")
            headers = {
                'Content-Length': f'{content_length}',
                'Content-Type': 'text/plain'
            }

            if cloud_provider == 'google':
                response = requests.put(new_url, data=payload, headers=headers)
                if not quiet:
                    print("submitted")
            else:
                # data = FormData()
                # for dk, dv in new_url.get('fields', {}).items():
                #     data.add_field(dk, dv)
                # data.add_field('file', open(filepath, 'rb'))
                response = requests.post(new_url.get('url'), data=new_url.get('fields', {}), files={'file': open(filepath, 'rb')})
                if not quiet:
                    print("submitted")


def upload_files_or_folder(args_folder, server, bucket, key, cloud_provider, org_config_id, args_files=None,
                           filepattern=None, filepattern_start=None, exclude_filepattern_end=None,
                           max_filesize=None, api_key=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False, req_metadata=None):
    global loop
    if args_folder:
        if max_connections and max_connections <= 1:
            upload_folder_sync(args_folder, server, bucket, subfolder=key, cloud_provider=cloud_provider,
                                org_config_id=org_config_id, filepattern=filepattern, filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key, max_connections=max_connections, filepatterns=filepatterns, quiet=quiet)
        else:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(
                upload_folder_async(args_folder, server, bucket, subfolder=key, cloud_provider=cloud_provider,
                                    org_config_id=org_config_id, filepattern=filepattern, filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key, max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs, req_metadata=req_metadata))
    elif args_files:
        cloud_upload(args_files, server, bucket, key, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)


async def upload_files_or_folder_async(args_folder, server, bucket, key, cloud_provider, org_config_id,
                                       filepattern=None, filepattern_start=None, exclude_filepattern_end=None,
                                       max_filesize=None, api_key=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, session=None):
    await upload_folder_async(args_folder, server, bucket, subfolder=key, cloud_provider=cloud_provider,
                        org_config_id=org_config_id, filepattern=filepattern, filepattern_start=filepattern_start,
                        exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key,
                        max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, session=session)
    # global loop
    # loop = asyncio.get_event_loop()
    # loop.run_until_complete(
    #     upload_folder_async(args_folder, server, bucket, subfolder=key, cloud_provider=cloud_provider,
    #                         org_config_id=org_config_id, filepattern=filepattern, filepattern_start=filepattern_start,
    #                         exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key,
    #                         max_connections=max_connections, filepatterns=filepatterns))
    # if args_folder:
    #     if max_connections and max_connections <= 1:
    #         upload_folder_sync(args_folder, server, bucket, subfolder=key, cloud_provider=cloud_provider,
    #                             org_config_id=org_config_id, filepattern=filepattern, filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key, max_connections=max_connections, filepatterns=filepatterns)
    #     else:
    # elif args_files:
    #     cloud_upload(args_files, server, bucket, key, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)


def check_pattern_match(pattern, path, default=True):
    if pattern is None:
        return default
    match = re.search(pattern, path)
    if match:
        return True
    else:
        return False



from ..executionhelpers import get_exec_info
from ..uploadhelpers import upload_files_or_folder, upload_files_or_folder_async


def upload_files_or_folder_for_exec(exec_id, miqa_server, args_folder=None, args_files=None, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False, req_metadata=None):
    """
      Uploads the test output files for each specified dataset, from the specified filepath to the corresponding Miqa execution.

      :param int exec_id: Miqa execution ID
      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str args_folder: Folder containing files to upload (Optional)
      :param list args_files: List of files to upload (Optional)
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :param int parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
      :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)
      :param dict req_metadata: Used for adding metadata to logs; only used for execution uploads associated with test chain runs. (Optional)
       """
    exec_info = get_exec_info(exec_id, miqa_server)
    bucket = exec_info.get('bucket')
    key = exec_info.get('key')
    cloud_provider = exec_info.get('cloud_provider', 'aws')
    org_config_id = exec_info.get('org_config_id')
    upload_files_or_folder(args_folder, miqa_server, bucket, key, cloud_provider, org_config_id, args_files,
                           filepattern=filepattern_end, filepattern_start=filepattern_start,
                           exclude_filepattern_end=exclude_filepattern_end,
                           max_filesize=max_filesize, api_key=api_key,
                           max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs, req_metadata=req_metadata)


async def upload_files_or_folder_for_exec_async(exec_id, server, args_folder=None, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, filepatterns=None, quiet=True, session=None):
    """
      Uploads the test output files from the specified filepath to the corresponding Miqa execution.

      :param exec_id: Execution ID
      :param server: Miqa server e.g. yourco.miqa.io
      :param args_folder: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param bool skip_completed: Whether to skip uploading files if there is already an execution with status Done, Failed, or Cancelled (Optional, default True)
      :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      """
    exec_info = get_exec_info(exec_id, server)
    bucket = exec_info.get('bucket')
    key = exec_info.get('key')
    cloud_provider = exec_info.get('cloud_provider', 'aws')
    org_config_id = exec_info.get('org_config_id')
    await upload_files_or_folder_async(args_folder, server, bucket, key, cloud_provider, org_config_id,
                                       filepattern=filepattern_end, filepattern_start=filepattern_start,
                                       exclude_filepattern_end=exclude_filepattern_end,
                                       max_filesize=max_filesize, api_key=api_key,
                                       max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, session=session)


import os

import requests

from ...utilities import get_json_or_error


def cloud_upload(files, server, bucket, subfolder="test", cloud_provider='google', org_config_id=None, api_key=None):
    urls = {}
    for filepath in files:
        url = upload_file(bucket, filepath, server, subfolder, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)
        urls[filepath] = url
    return urls


def upload_file(bucket, filepath, server, subfolder, cloud_provider='google', org_config_id=None, api_key=None):
    url = get_upload_url(bucket, filepath, server, subfolder, cloud_provider=cloud_provider, org_config_id=org_config_id, api_key=api_key)
    upload_file_to_url(filepath, url, cloud_provider=cloud_provider, api_key=api_key)
    return url


def upload_file_to_url(filepath, url, cloud_provider='google', api_key=None):
    if cloud_provider == 'google':
        content_length = os.path.getsize(filepath)
        payload = open(filepath, "rb")
        headers = {
            'Content-Length': f'{content_length}',
            'Content-Type': 'text/plain'
        }
        if api_key:
            headers['app_key'] = api_key
        response = requests.request("PUT", url, headers=headers, data=payload)
        print(response.text)
    else:
        with open(filepath, 'rb') as file_to_upload:
            files = {'file': (filepath, file_to_upload)}
            upload_response = requests.post(url['url'], data=url['fields'], files=files)
            print(upload_response.text)


def get_upload_url(bucket, filepath, server, subfolder, cloud_provider='google', org_config_id=None, api_key=None):
    remote_url = get_remote_url(bucket, filepath, server, subfolder, cloud_provider=cloud_provider, org_config_id=org_config_id)
    headers = {}
    if api_key:
        headers['app_key'] = api_key

    url = get_json_or_error(remote_url, headers=headers).get('url')
    return url


def get_remote_url(bucket, filepath, server, subfolder, cloud_provider='google', org_config_id=None, api_key=None):
    key = f"{subfolder}/{os.path.basename(filepath)}"
    remote_url = f'https://{server}/api/resumable_upload?bucket={bucket}&key={key}&cloud_provider={cloud_provider}'
    if api_key:
        remote_url += f"&app_key={api_key}"
    if org_config_id:
        remote_url += f"&org_config_id={org_config_id}"
    return remote_url


def get_user_log_post_url(server, run_id):
    return f"https://{server}/api/test_chain_run/{run_id}/add_user_logs"


import requests

from ..executionhelpers import complete_exec
from ..offlineexecution import upload_files_or_folder_for_exec
from ..triggertest_helpers import update_execution_start_time
from ...datamanagement.datasethelpers import get_ds_ids_by_name


def create_workflow_version(miqa_server, workflow_id, version_name, api_key):
    """
    Create a new manual-upload version for the specified workflow

    :param str miqa_server: Miqa server e.g. yourco.miqa.io
    :param int workflow_id: Miqa implemented workflow ID
    :param str version_name: Name for new version
    :param str api_key: API key

    :return: New workflow version ID
    :rtype: int
    """
    create_version_url = f"https://{miqa_server}/api/create_uploadable_version?pskel_id={workflow_id}&version_name={version_name}"
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    create_version_response = requests.get(create_version_url, headers=headers)
    if create_version_response.status_code == 200:
        create_version_response = create_version_response.json()
        wfv_id = create_version_response.get("workflowversion_id")
        return wfv_id
    else:
        raise Exception(f"Unable to create workflow version with name '{version_name}' for {workflow_id}: received an error response from the endpoint. Response: {create_version_response.text}")


def create_execution(miqa_server, workflowversion_id, dataset_id, org_config_id, api_key):
    """
    Create a new execution for the specified dataset on the specified workflow version

    :param str miqa_server: Miqa server e.g. yourco.miqa.io
    :param int workflowversion_id: Miqa workflow version ID
    :param int dataset_id: Miqa dataset ID
    :param int org_config_id: Miqa organization configuration ID
    :param str api_key: API key

    :return: New execution ID
    :rtype: int
    """
    create_execution_url = f"https://{miqa_server}/api/workflowversion/{workflowversion_id}/execute_on_dataset/{dataset_id}?org_config_id={org_config_id}"
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    create_execution_response = requests.get(create_execution_url, headers=headers)
    if create_execution_response.status_code == 200:
        create_execution_response = create_execution_response.json()
        exec_id = create_execution_response.get("exec_id")
        return exec_id
    else:
        raise Exception(f"Unable to create execution for {dataset_id} on {workflowversion_id}: received an error response from the endpoint at {create_execution_url}. Response: {create_execution_response.text}")


def create_and_upload_to_new_version_for_ds(miqa_server, version_name, dataset_name, pipeline_id, workflow_id, org_config_id, api_key, directory_containing_outputs, raise_on_fail=False, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False):
    """
      Create and execute a new version for an existing dataset, retrieved by name

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str version_name: Name for new version
      :param str dataset_name: Name for new dataset
      :param int pipeline_id: Miqa pipeline ID to add the datasets to
      :param int workflow_id: Miqa implemented workflow ID for the new version
      :param int org_config_id: Miqa organization configuration ID
      :param str api_key: API key
      :param str directory_containing_outputs: Folder containing files to upload (Optional)
      :param bool raise_on_fail: Whether to raise an exception if the file upload fails. Default: False, i.e. the execution will be failed but an exception won't be raised (Optiona)
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :param int parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
      :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)

    """
    exec_id = create_and_execute_for_new_version_for_ds(miqa_server, version_name, dataset_name, pipeline_id, workflow_id, org_config_id, api_key)
    upload_and_complete_execution(miqa_server, exec_id, api_key, directory_containing_outputs, raise_on_fail=raise_on_fail, filepattern_end=filepattern_end, filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs)
    return exec_id


def upload_and_complete_execution(miqa_server, exec_id, api_key, directory_containing_outputs, raise_on_fail=False, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False):
    """
      Uploads the test output files for each specified dataset, from the specified filepath to the corresponding Miqa execution, and then marks that execution as complete.

      :param int exec_id: Miqa execution ID
      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str directory_containing_outputs: Folder containing files to upload (Optional)
      :param bool raise_on_fail: Whether to raise an exception if the file upload fails. Default: False, i.e. the execution will be failed but an exception won't be raised (Optiona)
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :param int parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
      :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)
       """
    update_execution_start_time(miqa_server, exec_id, api_key=api_key, quiet=quiet)
    failed = False
    try:
        upload_files_or_folder_for_exec(exec_id, miqa_server, directory_containing_outputs, args_files=None,
                                        filepattern_end=filepattern_end,
                                        filepattern_start=filepattern_start,
                                        exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize,
                                        api_key=api_key,
                                        max_connections=max_connections, filepatterns=filepatterns, quiet=quiet,
                                        per_request_timeout=per_request_timeout,
                                        parallel_file_upload=parallel_file_upload,
                                        detailed_file_logs=detailed_file_logs)

    except Exception as e2:
        if raise_on_fail:
            raise
        failed = True
        if not quiet:
            print(
                f"Failing execution {exec_id} due to unhandled exception (see below). Will mark execution as failed and continue any remaining samples.")
            print(e2)
            complete_exec(exec_id, miqa_server, quiet=quiet, api_key=api_key, failed=failed)
    complete_exec(exec_id, miqa_server, quiet=quiet, api_key=api_key, failed=failed)


def create_and_execute_for_new_version_for_ds(miqa_server, version_name, dataset_name, pipeline_id, workflow_id, org_config_id, api_key):
    """
      Create and execute a new version for an existing dataset, retrieved by name

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str version_name: Name for new version
      :param str dataset_name: Name for new dataset
      :param int pipeline_id: Miqa pipeline ID to add the datasets to
      :param int workflow_id: Miqa implemented workflow ID for the new version
      :param int org_config_id: Miqa organization configuration ID
      :param str api_key: API key
    """
    ds_ids_by_name = get_ds_ids_by_name(miqa_server, pipeline_id, api_key)
    ds_id = ds_ids_by_name.get("data",{}).get(dataset_name)
    if not ds_id:
        raise Exception(f"Could not find a dataset by the name '{dataset_name}' for the provided pipeline (id={pipeline_id}). Do you have a typo or the wrong pipeline?")
    wfv_id = create_workflow_version(miqa_server, workflow_id, version_name, api_key)
    exec_id = create_execution(miqa_server, wfv_id, ds_id, org_config_id, api_key)
    return exec_id



import requests

from ...utilities import get_json_or_error


def get_ds_ids_by_name(miqa_server, pipeline_id, api_key=None):
    """
      Gets a lookup of dataset names and corresponding IDs for a particular pipeline

      :param miqa_server: Miqa server e.g. yourco.miqa.io
      :param pipeline_id: Miqa pipeline ID to get dataset info for
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: A lookup of dataset information
      :rtype: dict
    """
    info = get_ds_ids_by_name_internal(miqa_server, pipeline_id, api_key=api_key)
    return info


def get_ds_ids_by_name_internal(miqa_server, pipeline_id, api_key=None):
    remote_url = f'https://{miqa_server}/api/get_ds_ids_by_name?pipeline_id={pipeline_id}'
    headers = {}
    if api_key:
        headers['app_key'] = api_key
    info = get_json_or_error(remote_url, headers=headers)
    return info


def tag_ds_ids_by_name(miqa_server, pipeline_id, name, tag, api_key=None):
    """
      Tags dataset(s) matching the name provided with the tag provided

      :param miqa_server: Miqa server e.g. yourco.miqa.io
      :param pipeline_id: Miqa pipeline ID to get dataset info for
      :param str name: Dataset name corresponding to the dataset to apply the tag to
      :param str tag: Tag to use for the dataset
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: Response from the API call
      :rtype: object
    """
    info = get_ds_ids_by_name_internal(miqa_server, pipeline_id)
    source_ids_lookup = info.get('data',{})
    headers = {}
    if api_key:
        headers['app_key'] = api_key

    if source_ids_lookup and source_ids_lookup.get(name):
        tag_url = f"https://{miqa_server}/api/batch_tag_datasources/{tag}?inline=1&source_ids={source_ids_lookup.get(name)}"
        resp = requests.get(tag_url, headers=headers)
        return resp
    else:
        raise Exception("No matching datasource IDs found")


def tag_ds_ids_by_names(miqa_server, pipeline_id, names, tag, raise_if_names_not_found=False, api_key=None):
    """
      Tags dataset(s) matching the names provided with the tag provided

      :param miqa_server: Miqa server e.g. yourco.miqa.io
      :param pipeline_id: Miqa pipeline ID to get dataset info for
      :param list name: Dataset name corresponding to the dataset to apply the tag to
      :param str tag: Tag to use for the dataset
      :param bool raise_if_names_not_found: Whether to raise an exception if no datasets are found matching one of the provided names
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: Response from the API call
      :rtype: object
    """
    info = get_ds_ids_by_name_internal(miqa_server, pipeline_id)
    source_ids_lookup = info.get('data',{})
    source_ids_for_names = [source_ids_lookup.get(name) for name in names]
    source_ids_for_names = [str(i) for i in source_ids_for_names if i]

    headers = {}
    if api_key:
        headers['app_key'] = api_key

    if source_ids_for_names and (not raise_if_names_not_found or len(source_ids_for_names) == len(names)):
        tag_url = f"https://{miqa_server}/api/batch_tag_datasources/{tag}?inline=1&source_ids={','.join(source_ids_for_names)}"
        resp = requests.get(tag_url, headers=headers)
        return resp
    else:
        raise Exception("Matching datasource IDs not found")


def batch_create_datasets_and_tag_or_group(miqa_server, org_id, pipeline_id, new_ds_names, tags, dsg_name=None, api_key=None):
    """
      Create stand-in datasets (i.e. no data upload) and tag them

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int org_id: Miqa organization ID
      :param int pipeline_id: Miqa pipeline ID to add the datasets to
      :param list new_ds_names: Names for new datasets
      :param list tags: Tags to use for the dataset
      :param str dsg_name: Name to use to create a dataset group, if wanting to create a group (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :return: Response from the API call
      :rtype: Response
    """
    tag_url = f"https://{miqa_server}/api/batch_create_dataset_and_tag?org_id={org_id}&pipeline_id={pipeline_id}&names={','.join(new_ds_names)}"

    if tags:
        tag_url += f"&tags={','.join(tags)}"
    if dsg_name:
        tag_url += f"&create_dsg={dsg_name}"

    headers = {}
    if api_key:
        headers['app_key'] = api_key

    resp = requests.get(tag_url, headers=headers)
    return resp


def add_truths_to_dataset(miqa_server, dataset_id, pipeline_id, truth_file_path, api_key):
    """
      Create a single stand-in dataset (i.e. no data upload) and optionally tag it

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int dataset_id: Miqa dataset ID to add the truths to
      :param int pipeline_id: Miqa pipeline ID corresponding to the dataset and truth type
      :param str truth_file_path: Filepath pointing to the truth TSV file. Should be a tab-delimited file with columns and headers corresponding to your pipeline's truth type, e.g. "chr   pos ref alt"
      :param str api_key: API key
    """
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    upload_truths_url = f"https://{miqa_server}/api/datasource/{dataset_id}/upload_truths?pipeline_id={pipeline_id}&inline=1"
    response = requests.post(upload_truths_url, files={'file': open(truth_file_path, 'rb')}, data={}, headers=headers)
    if response.status_code != 200:
        raise Exception(f"Unable to add truths to dataset {dataset_id}: received an error response from the server at url {upload_truths_url}. Response: {response.text}")
    return response


def create_dataset_with_name(miqa_server, org_id, pipeline_id, dataset_name, api_key, tags=None):
    """
      Create a single stand-in dataset (i.e. no data upload) and optionally tag it

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int org_id: Miqa organization ID
      :param int pipeline_id: Miqa pipeline ID to add the datasets to
      :param str dataset_name: Name for new dataset
      :param str api_key: API key
      :param list tags: Tags to use for the dataset (optional)

      :return: New Dataset ID
      :rtype: int
    """
    create_ds_response = batch_create_datasets_and_tag_or_group(miqa_server, org_id, pipeline_id,
                                                                new_ds_names=[dataset_name],
                                                                tags=tags if tags else [], api_key=api_key)  # Add tags if desired
    if create_ds_response.status_code == 200:
        create_ds_response = create_ds_response.json()
        ds_id_lookup = create_ds_response.get("data")
        ds_id = ds_id_lookup.get(dataset_name)
        return ds_id
    else:
        raise Exception(f"Unable to create dataset with name '{dataset_name}': received an error response from the endpoint. Response: {create_ds_response.text}")


def create_dataset_and_add_truths(miqa_server, org_id, pipeline_id, dataset_name, truth_file_path, api_key, tags=None):
    """
      Create a single stand-in dataset (i.e. no input file upload), add truth data, and optionally tag it

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int org_id: Miqa organization ID
      :param int pipeline_id: Miqa pipeline ID corresponding to the dataset and truth type
      :param str dataset_name: Name for new dataset
      :param str truth_file_path: Filepath pointing to the truth TSV file. Should be a tab-delimited file with columns and headers corresponding to your pipeline's truth type, e.g. "chr   pos ref alt"
      :param str api_key: API key
      :param list tags: Tags to use for the dataset (optional)

      :return: New Dataset ID
      :rtype: int
    """
    dataset_id = create_dataset_with_name(miqa_server, org_id, pipeline_id, dataset_name, api_key, tags=tags)
    add_truths_to_dataset(miqa_server, dataset_id, pipeline_id, truth_file_path, api_key)
    return dataset_id



import requests

from ...utilities import get_json_or_error


def complete_exec(exec_id, facet_server=None, quiet=True, api_key=None, failed=False):
    if not quiet:
        print(f"Completing execution {exec_id}")
    remote_url = f'https://{facet_server}/api/execution/{exec_id}/mark_{"complete" if not failed else "failed"}'
    headers = {}
    if api_key:
        headers['app_key'] = api_key
    info = get_json_or_error(remote_url, headers=headers)
    if not quiet:
        print(info.get('message'))
    return info


def get_exec_info(exec_id, facet_server=None, api_key=None):
    remote_url = f'https://{facet_server}/api/execution/{exec_id}'
    headers = {}
    if api_key:
        headers['app_key'] = api_key
    info = get_json_or_error(remote_url, headers=headers)
    return info


def get_exec_url(exec_id, facet_server=None):
    return f'https://{facet_server}/execution/{exec_id}'



from argparse import ArgumentParser

import asyncio

from ..executionhelpers import complete_exec
from ..offlineexecution import upload_files_or_folder_for_exec, upload_files_or_folder_for_exec_async
from ..triggertest_helpers import trigger_miqa_test, get_tcr_info_json, update_execution_start_time


def trigger_test_and_upload_single_results(miqa_server, trigger_id, version_name, directory_containing_outputs, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None):
    """
      Triggers an offline execution of the specified Miqa trigger ID for the specified version, and then uploads the test output files, from the specified filepath to the corresponding Miqa execution.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str trigger_id: Miqa test trigger ID e.g. ABC123
      :param str version_name: Name representing the version that was being tested/generated the outputs being uploaded, e.g. a commit ID ('abc123') or version number ('v1.2.3')
      :param str directory_containing_outputs: Directory filepath containing the test outputs/results to be uploaded
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param bool skip_completed: Whether to skip uploading files if there is already an execution with status Done (Optional, default True)
      :param list filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :return: Test chain run ID
      :rtype: int
      """
    run_id = trigger_miqa_test(miqa_server, trigger_id, version_name)
    get_info_response_json = get_tcr_info_json(miqa_server, run_id, directory_containing_outputs)
    exec_id = get_info_response_json.get("exec_id")
    upload_files_or_folder_for_exec(exec_id, miqa_server, directory_containing_outputs, None, filepattern_end=filepattern_end,
                                    filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, api_key=api_key,
                                    max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout)
    complete_exec(exec_id, miqa_server, quiet=quiet)
    return run_id


def trigger_test_and_upload_by_dsid(miqa_server, trigger_id, version_name, ds_lookup, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, api_key=None, max_connections=None, skip_completed=True, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False, treat_failed_as_completed=True, treat_cancelled_as_completed=True):
    """
      Triggers an offline execution of the specified Miqa trigger ID for the specified version, and then uploads the test output files for each specified dataset, from the specified filepath to the corresponding Miqa execution.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param str trigger_id: Miqa test trigger ID e.g. ABC123
      :param str version_name: Name representing the version that was being tested/generated the outputs being uploaded, e.g. a commit ID ('abc123') or version number ('v1.2.3')
      :param dict ds_lookup: Dictionary of dataset IDs (int) and their locations (filepath string), e.g. {DS_ID_1:'OUTPUT_LOCATION_FOR_DS_ID_1',DS_ID_2:'OUTPUT_LOCATION_FOR_DS_ID_2'}
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param bool skip_completed: Whether to skip uploading files if there is already an execution with status Done, Failed, or Cancelled (Optional, default True)
      :param list filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :param int parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
      :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)
      :param bool treat_failed_as_completed: Whether to treat executions with status 'failed' as completed, i.e. skip uploading to them. Default True. (Optional)
      :param bool treat_cancelled_as_completed: Whether to treat executions with status 'cancelled' as completed, i.e. skip uploading to them. Default True. (Optional)
      :return: Test chain run ID
      :rtype: int
      """
    run_id = trigger_miqa_test(miqa_server, trigger_id, version_name, api_key=api_key)
    upload_to_test_by_dsid(run_id, miqa_server, ds_lookup, filepattern_end, filepattern_start, exclude_filepattern_end, api_key,
                           max_connections, max_filesize, skip_completed=skip_completed, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs, treat_failed_as_completed=treat_failed_as_completed, treat_cancelled_as_completed=treat_cancelled_as_completed)
    return run_id


def upload_to_test_by_dsid(run_id, miqa_server, ds_lookup, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, api_key=None,
                           max_connections=None, max_filesize=None, skip_completed=True, filepatterns=None, quiet=True, halt_on_upload_failure=False, halt_on_general_failure=False, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False, treat_failed_as_completed=True, treat_cancelled_as_completed=True):
    """
      Uploads the test output files for each specified dataset, from the specified filepath to the corresponding Miqa execution.

      :param int run_id: Test chain run ID
      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param dict ds_lookup: Dictionary of dataset IDs (int) and their locations (filepath string), e.g. {DS_ID_1:'OUTPUT_LOCATION_FOR_DS_ID_1',DS_ID_2:'OUTPUT_LOCATION_FOR_DS_ID_2'}
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param bool skip_completed: Whether to skip uploading files if there is already an execution with status Done, Failed, or Cancelled (Optional, default True)
      :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      :param bool halt_on_upload_failure: Whether to halt all processing if there is an error (e.g. timeout) in the upload. Default False. (Optional)
      :param bool halt_on_general_failure: Whether to halt all processing if there is an error (e.g. timeout) other than in the upload. Default False. (Optional)
      :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
      :param int parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
      :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)
      :param bool treat_failed_as_completed: Whether to treat executions with status 'failed' as completed, i.e. skip uploading to them. Default True. (Optional)
      :param bool treat_cancelled_as_completed: Whether to treat executions with status 'cancelled' as completed, i.e. skip uploading to them. Default True. (Optional)
       """
    completed_statuses = ["done"]
    if treat_failed_as_completed:
        completed_statuses.append("failed")
    if treat_cancelled_as_completed:
        completed_statuses.append("cancelled")
    for ds_id, directory_containing_outputs in ds_lookup.items():
        try:
            get_info_response_json = get_tcr_info_json(miqa_server, run_id, directory_containing_outputs, ds_id=ds_id)
            exec_id = get_info_response_json.get("exec_id")
            failed = False
            req_metadata = {"tcr_id": run_id, "exec_id": exec_id}
            try:
                if skip_completed:
                    if "exec_status" in get_info_response_json:
                        if get_info_response_json.get("exec_status").lower() in completed_statuses:
                            if not quiet:
                                print(f"Skipping upload for {exec_id}: status is already {get_info_response_json.get('exec_status')}")
                            continue
                # Set updated start time
                update_execution_start_time(miqa_server, exec_id, api_key=api_key, quiet=quiet)

                upload_files_or_folder_for_exec(exec_id, miqa_server, directory_containing_outputs, None,
                                            filepattern_end=filepattern_end,
                                            filepattern_start=filepattern_start,
                                            exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize,
                                            api_key=api_key,
                                            max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs, req_metadata=req_metadata)
            except Exception as e2:
                failed = True
                if not quiet:
                    print(f"Failing {exec_id} for {ds_id} due to unhandled exception (see below). Will mark execution as failed and continue any remaining samples.")
                    print(e2)
                if halt_on_upload_failure:
                    complete_exec(exec_id, miqa_server, quiet=quiet, api_key=api_key, failed=failed)
                    raise
            complete_exec(exec_id, miqa_server, quiet=quiet, api_key=api_key, failed=failed)
        except Exception as e:
            print(f"Failed upload for {ds_id}. See exception below.")
            print(e)
            if halt_on_general_failure:
                raise


def upload_to_test_by_dsid_async(run_id, miqa_server, ds_lookup, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, api_key=None,
                           max_connections=None, max_filesize=None, skip_completed=True, filepatterns=None, quiet=True):
    """
      Uploads the test output files for each specified dataset, from the specified filepath to the corresponding Miqa execution.

      :param int run_id: Test chain run ID
      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param dict ds_lookup: Dictionary of dataset IDs (int) and their locations (filepath string), e.g. {DS_ID_1:'OUTPUT_LOCATION_FOR_DS_ID_1',DS_ID_2:'OUTPUT_LOCATION_FOR_DS_ID_2'}
      :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
      :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
      :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
      :param str api_key: API key, if required by your Miqa instance (Optional)
      :param int max_connections: Maximum number of connections for parallel upload (Optional)
      :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
      :param bool skip_completed: Whether to skip uploading files if there is already an execution with status Done, Failed, or Cancelled (Optional, default True)
      :param list filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
      :param bool quiet: Whether to skip logging. Default True. (Optional)
      """
    global loop
    loop = asyncio.get_event_loop()
    exec_id_lookup = {}
    for ds_id, directory_containing_outputs in ds_lookup.items():
        get_info_response_json = get_tcr_info_json(miqa_server, run_id, directory_containing_outputs, ds_id=ds_id)
        exec_id = get_info_response_json.get("exec_id")
        exec_id_lookup[ds_id] = exec_id
        if skip_completed:
            if "exec_status" in get_info_response_json:
                if get_info_response_json.get("exec_status").lower() in ["done", "failed", "cancelled"]:
                    if not quiet:
                        print(f"Skipping upload for {exec_id}: status is already Done")
                    continue
        loop.run_until_complete(
        upload_files_or_folder_for_exec_async(exec_id, miqa_server, directory_containing_outputs,
                                              filepattern_end=filepattern_end,
                                              filepattern_start=filepattern_start,
                                              exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize,
                                              api_key=api_key,
                                              max_connections=max_connections, filepatterns=filepatterns))

    for ds_id, directory_containing_outputs in ds_lookup.items():
        complete_exec(exec_id_lookup[ds_id], miqa_server, quiet=quiet)


if __name__ == '__main__':
    parser = ArgumentParser(description='A command line tool for interacting with the Miqa API')
    parser.add_argument('--trigger_id', type=str, default=None, help='Trigger ID in Miqa')
    parser.add_argument('--server', type=str, default=None, help='Miqa Server URL')
    parser.add_argument('--version_name', type=str, default=None, help='Version Name to create (e.g. MyPipeline v1.0, or commit ID e.g. abc123de')
    parser.add_argument('--directory', type=str, default=None, help='Path to local directory containing files to upload')
    args = parser.parse_args()

    trigger_test_and_upload_single_results(args.server, args.trigger_id, args.version_name, args.directory)




from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='miqatools',
    version='1.1.10rc2',
    license='GPLv3',
    author="Gwenn Berry",
    author_email='gwenn@magnalabs.co',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://www.magnalabs.co',
    keywords='miqa',
    install_requires=[
          'asyncio', 'aiohttp',
      ],
    long_description=long_description,
    long_description_content_type='text/x-rst'
)


import requests

from ..adhoc_execution import create_and_upload_to_new_version_for_ds
from ...datamanagement.datasethelpers import create_dataset_and_add_truths, get_ds_ids_by_name


def get_organization_configuration(miqa_server, org_config_id, api_key):
    """
    Create new datasets with reference/truth, and upload results to evaluate accuracy

    :param str miqa_server: Miqa server e.g. yourco.miqa.io
    :param int org_config_id: Miqa organization configuration ID
    :param str api_key: API key
    """

    url = f"https://{miqa_server}/api/organization_configuration/{org_config_id}"
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return response.json()
    else:
        raise Exception(f"Unable to get organization configuration information. Response: {response.text}")


def get_implemented_workflow_info(miqa_server, workflow_id, api_key):
    """
    Create new datasets with reference/truth, and upload results to evaluate accuracy

    :param str miqa_server: Miqa server e.g. yourco.miqa.io
    :param int workflow_id: Miqa implemented workflow ID
    :param str api_key: API key
    """
    url = f"https://{miqa_server}/api/implemented_workflow/{workflow_id}"
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    response = requests.get(url, headers=headers)
    if response.status_code == 200:
        return response.json()
    else:
        raise Exception(f"Unable to get implemented workflow information. Response: {response.text}")


def verify_no_ds_with_name(miqa_server, pipeline_id, dataset_name, api_key):
    ds_ids_by_name = get_ds_ids_by_name(miqa_server, pipeline_id, api_key)
    ds_id = ds_ids_by_name.get("data",{}).get(dataset_name)
    if ds_id:
        raise Exception(f"Already found a dataset with name '{dataset_name}' for the provided pipeline (id={pipeline_id}). Please provide a unique name.")


def create_dataset_with_truths_and_check_results(miqa_server, org_config_id, workflow_id, dataset_name, truth_file_path, version_name, directory_containing_outputs, api_key, tags=None, recall_threshold=0.9, precision_threshold=0.9, raise_on_fail=False, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False):
    """
    Create new datasets with reference/truth, and upload results to evaluate accuracy

    :param str miqa_server: Miqa server e.g. yourco.miqa.io
    :param int org_config_id: Miqa organization configuration ID
    :param int workflow_id: Miqa implemented workflow ID
    :param str dataset_name: Name for new dataset
    :param version_name: Name representing the version that was being tested/generated the outputs being uploaded, e.g. a commit ID ('abc123') or version number ('v1.2.3')
    :param str truth_file_path: Filepath pointing to the truth TSV file. Should be a tab-delimited file with columns and headers corresponding to your pipeline's truth type, e.g. "chr   pos ref alt"
    :param str api_key: API key
    :param list tags: Tags to use for the dataset (optional)
    :param str directory_containing_outputs: Folder containing files to upload (Optional)
    :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
    :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
    :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
    :param int max_connections: Maximum number of connections for parallel upload (Optional)
    :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
    :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
    :param bool quiet: Whether to skip logging. Default True. (Optional)
    :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
    :param bool parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
    :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)

    :return: Test result response json
    :rtype: Response
    """
    # Can we get pipeline for workflow?
    # Can we get org for workflow?
    implemented_workflow_info = get_implemented_workflow_info(miqa_server, workflow_id, api_key)
    pipeline_id = implemented_workflow_info.get("pipeline_id")
    org_config_info = get_organization_configuration(miqa_server, org_config_id, api_key)
    org_id = org_config_info.get("organization_id")

    verify_no_ds_with_name(miqa_server, pipeline_id, dataset_name, api_key)

    ds_id = create_dataset_and_add_truths(miqa_server, org_id, pipeline_id, dataset_name, truth_file_path, api_key, tags=tags)
    return check_results_for_dataset_with_truths(miqa_server, org_config_id, workflow_id, dataset_name, version_name, directory_containing_outputs, api_key, recall_threshold=recall_threshold, precision_threshold=precision_threshold, raise_on_fail=raise_on_fail, filepattern_end=filepattern_end, filepattern_start=filepattern_start, exclude_filepattern_end=exclude_filepattern_end, max_filesize=max_filesize, max_connections=max_connections, filepatterns=filepatterns, quiet=quiet, per_request_timeout=per_request_timeout, parallel_file_upload=parallel_file_upload, detailed_file_logs=detailed_file_logs)


def check_results_for_dataset_with_truths(miqa_server, org_config_id, workflow_id, dataset_name, version_name, directory_containing_outputs, api_key, recall_threshold=0.9, precision_threshold=0.9, raise_on_fail=False, filepattern_end=None, filepattern_start=None, exclude_filepattern_end=None, max_filesize=None, max_connections=None, filepatterns=None, quiet=True, per_request_timeout=None, parallel_file_upload=True, detailed_file_logs=False):
    """
       Upload results to evaluate accuracy on dataset with existing truth info

       :param str miqa_server: Miqa server e.g. yourco.miqa.io
       :param int org_config_id: Miqa organization configuration ID
       :param int workflow_id: Miqa implemented workflow ID
       :param str dataset_name: Name for new dataset
       :param version_name: Name representing the version that was being tested/generated the outputs being uploaded, e.g. a commit ID ('abc123') or version number ('v1.2.3')
       :param str api_key: API key
       :param str directory_containing_outputs: Folder containing files to upload (Optional)
       :param str filepattern_end: Pattern to match in the file suffix, e.g. '.bam' (Optional)
       :param str filepattern_start: Pattern to match in the file prefix, e.g. 'Analysis/' (Optional)
       :param str exclude_filepattern_end: Pattern to exclude in the file suffix, e.g. '.bam' (Optional)
       :param int max_connections: Maximum number of connections for parallel upload (Optional)
       :param int max_filesize: Maximum filesize to upload, i.e. if avoiding very large files (Optional)
       :param filepatterns: List of regex strings representing file patterns to match, e.g. ['*.csv','*.vcf'] (Optional)
       :param bool quiet: Whether to skip logging. Default True. (Optional)
       :param int per_request_timeout: Timeout for individual requests made during uploading. Default None. (Optional)
       :param bool parallel_file_upload: Whether to allow uploads of each file to proceed in parallel. Default True. (Optional)
       :param bool detailed_file_logs: If not in quiet mode, whether to show detailed logs of each file uploaded. Default False. (Optional)

       :return: Test result response json
       :rtype: Response
    """
    implemented_workflow_info = get_implemented_workflow_info(miqa_server, workflow_id, api_key)
    pipeline_id = implemented_workflow_info.get("pipeline_id")
    exec_id = create_and_upload_to_new_version_for_ds(miqa_server, version_name, dataset_name, pipeline_id, workflow_id,
                                                      org_config_id if org_config_id else org_id, api_key,
                                                      directory_containing_outputs, raise_on_fail=raise_on_fail,
                                                      filepattern_end=filepattern_end,
                                                      filepattern_start=filepattern_start,
                                                      exclude_filepattern_end=exclude_filepattern_end,
                                                      max_filesize=max_filesize, max_connections=max_connections,
                                                      filepatterns=filepatterns, quiet=quiet,
                                                      per_request_timeout=per_request_timeout,
                                                      parallel_file_upload=parallel_file_upload,
                                                      detailed_file_logs=detailed_file_logs)
    assertions_json = get_empty_assertions_json()
    add_assertion_obj(assertions_json, get_standard_assertion("recall", threshold=recall_threshold))
    add_assertion_obj(assertions_json, get_standard_assertion("precision", threshold=precision_threshold))
    test_result = perform_adhoc_test_on_single_exec(miqa_server, exec_id, assertions_json, api_key)
    return test_result


def perform_adhoc_test_on_single_exec(miqa_server, exec_id, assertions_json, api_key, allow_validation_failures=False):
    """
      Evaluate test assertions for a single execution in Miqa.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int exec_id: Miqa execution ID
      :param dict assertions_json: Assertions JSON in same format as used for test block assertions
      :param str api_key: API key
      :param bool allow_validation_failures: When checking whether assertion types are valid for single-execution tests, whether to allow the test to proceed (use value True) or raise an exception (use value False). Default: False (Optional)

      :return: Test results (pass/fail) per category and check
      :rtype: dict
    """
    for k,v in assertions_json.get("assertions", {}).items():
        if k in ["concordance", "filecompare"]:
            message = f"Assertions of type {k} are not valid when performing single-execution tests."
            if not allow_validation_failures:
                raise Exception(message + " You may still perform this test by using parameter allow_validation_failures=True.")
            else:
                print(f"WARNING: {message}")
        
    return perform_adhoc_test_on_paired_execs(miqa_server, exec_id, exec_id, assertions_json, api_key)


def perform_adhoc_test_on_paired_execs(miqa_server, baseline_id, test_id, assertions_json, api_key):
    """
      Evaluate test assertions for a pair of executions in Miqa.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param int baseline_id: Baseline execution IDs
      :param int test_id: Baseline execution IDs
      :param dict assertions_json: Assertions JSON in same format as used for test block assertions
      :param str api_key: API key

      :return: Test results (pass/fail) per category and check
      :rtype: dict
    """
    return perform_adhoc_test(miqa_server, [baseline_id], [test_id], assertions_json, api_key)


def perform_adhoc_test(miqa_server, baseline_ids, test_ids, assertions_json, api_key):
    """
      Evaluate test assertions for a set of baseline and test executions in Miqa.

      :param str miqa_server: Miqa server e.g. yourco.miqa.io
      :param list baseline_ids: Baseline execution IDs
      :param list test_ids: Baseline execution IDs
      :param dict assertions_json: Assertions JSON in same format as used for test block assertions
      :param str api_key: API key

      :return: Test results (pass/fail) per category and check
      :rtype: dict
    """
    url = f"https://{miqa_server}/api/adhoc_test_results?inline=1&exec_ids_baseline={','.join([str(i) for i in baseline_ids])}&exec_ids_test={','.join([str(i) for i in test_ids])}"
    headers = {
        "accept": "application/json",
        "app_key": api_key
    }
    response = requests.post(url, json=assertions_json, headers=headers)
    if response.status_code == 200:
        return response.json()
    else:
        raise Exception(f"Unable to perform adhoc test: received an error response from the endpoint at {url}. Response: {response.text}")


def get_empty_assertions_json():
    """
      Create an empty assertions json to populate with standard assertions or custom assertions of various categories.
      :return: Empty assertions JSON
      :rtype: dict
    """
    return {"assertions": {}}


def get_standard_assertion(assertion_type, threshold=None):
    """
      Get a standard assertion as a starting point

      :param str assertion_type: Assertion type, one of ['concordance', 'recall', 'precision', 'vcf_filecompare', 'csv_filecompare']
      :param int org_id: Miqa organization ID
      :param int pipeline_id: Miqa pipeline ID to add the datasets to
      :param str dataset_name: Name for new dataset
      :param str api_key: API key
      :param list tags: Tags to use for the dataset (optional)

      :return: Assertion object ({"check":..., "category":...})
      :rtype: dict
    """

    assertion_lookup = {
        "concordance": {"check": {
                    "name": "Result-based concordance w/ Previous",
                    "check_type": "concordance",
                    "failtype": "fail",
                    "versions": [
                        -1
                    ],
                    "threshold": 0 if not threshold else threshold,
                }, "category": "concordance"},
        "recall": {"check": {
                    "name": "Overall recall",
                    "stat": "recall",
                    "check_type": "accuracy",
                    "failtype": "fail",
                    "relationship": "lt",
                    "threshold": 0 if not threshold else threshold,
                    "threshold_type": "percent",
                    "raw_tolerance": 0
                  }, "category": "accuracy"},
        "precision": {"check": {
                    "name": "Overall precision",
                    "stat": "precision",
                    "check_type": "accuracy",
                    "failtype": "fail",
                    "relationship": "lt",
                    "threshold": 0 if not threshold else threshold,
                    "threshold_type": "percent",
                    "raw_tolerance": 0
                  }, "category": "accuracy"},
        "vcf_filecompare": {"check":{
                    "name": "Line-by-line VCF Comparison w/ Previous",
                    "check_type": "filecompare",
                    "failtype": "fail",
                    "min_count": 1,
                    "delimiter": "\t",
                    "pct_diff_threshold": 0,
                    "skiprules": [],
                    "versions": [
                      -1
                    ],
                    "file_rules": {
                      "pattern": ".*vcf$"
                    },
                    "id": "blurb_vcf_1"
      }, "category":"filecompare"},
        "csv_filecompare": {"check":{
                    "name": "Line-by-line CSV Comparison w/ Previous",
                    "check_type": "filecompare",
                    "failtype": "fail",
                    "min_count": 1,
                    "delimiter": ",",
                    "pct_diff_threshold": 0,
                    "skiprules": [],
                    "versions": [
                      -1
                    ],
                    "file_rules": {
                      "pattern": ".*csv$"
                    },
                    "id": "blurb_vcf_1"
      }, "category":"filecompare"}
    }
    return assertion_lookup.get(assertion_type)


def add_assertion_obj(assertions_json, new_assertion_obj):
    """
    Add an assertion to the assertions JSON by passing a check and a category

    :param dict assertions_json: Assertions JSON to add onto
    :param dict new_assertion_obj: A JSON object containing keys {'check': {new_assertion_check...}, 'category':...}
    """
    add_assertion(assertions_json, new_assertion_obj.get('check'), new_assertion_obj.get('category'))


def add_assertion(assertions_json, new_assertion_check, new_assertion_category):
    """
    Add an assertion to the assertions JSON by passing a check and a category

    :param dict assertions_json: Assertions JSON to add onto
    :param dict new_assertion_check: New assertion "check" in JSON form
    :param str new_assertion_category: New assertion category, e.g. 'concordance' or 'accuracy'
    """
    if new_assertion_category not in assertions_json.get("assertions", {}):
        assertions_json["assertions"][new_assertion_category] = {"report":True, "checks": []}
    assertions_json["assertions"][new_assertion_category]["checks"].append(new_assertion_check)


import requests


def get_json_or_error(url, headers, suppress_error_text=False):
    response = requests.get(url, headers=headers)

    if response and response.ok:
        return response.json()
    else:
        if not suppress_error_text:
            print(response.text)
        raise Exception(f"Error calling endpoint {url}: {response.status_code} {response.text if not 'DOCTYPE html' in response.text else '(See message above)'}")